/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <cerrno>
#include <cstdlib>
#include <iostream>
#include <string>

#include <boost/asio.hpp>
#include <client.hpp>

#include "debug.hpp"

class ctx {

public:

	std::string host;
	uint16_t port;

	boost::asio::ip::tcp::iostream fds;

	ctx()
	: ctx( "127.0.0.1", 1234 )
	{
	}

	ctx( const std::string & host, const uint16_t port )
	:
		host( host ),
		port( port ),
		fds( host, std::to_string( port ) )
	{
		if ( ! fds ) {
			E() << "failed to connect to " << host << ":" << port << ": " << fds.error().message() << std::endl;
		} else {
			D() << "opened socket to " << host << ":" << port << std::endl;
			fds.unsetf( std::ios::skipws );
		}
	}

	~ctx() {
		close();
	}

	void close() {
		fds.close();
		D() << "closed socket to " << host << ":" << port << std::endl;
	}
};

/*
static void usage( const std::string & progname ) {
	std::cout
		<< "usage(): "
		<< progname
		<< " <ip>"
		<< " <port>"
		<< std::endl
		;
}
*/

int main( int argc, char *argv[] ) {
	std::string line;

	ctx ctx;

	static const std::vector<std::string> query {
		"$qSupported:multiprocess+;swbreak+;hwbreak+;qRelocInsn+#c9",
		"$Hg0#df",
		"$qXfer:features:read:target.xml:0,ffb#79",
		"$qXfer:features:read:arm-core.xml:0,ffb#08",
		"$qTStatus#49",
		"$?#3f",
		"$qfThreadInfo#bb",
		"$qsThreadInfo#c8",
		"$qAttached#8f",
		"$Hc-1#09",
		"$qOffsets#4b",
		"$g#67",
		"$m48,4#39",
		"$qSymbol::#5b",
		"$D#44"
	};

	if ( ! ctx.fds ) {
		return 1;
	}

//	std::stringstream ss;
//	ss << "+$PacketSize=1000;qXfer:features:read+#cc";

	// XXX: @CF: 20170916: the call below hangs
	// XXX: my suspicion is that this is because SO_TIMEOUT is not set
	// Boost ASIO suggests wrapping istream in a class and passing a class around.
	// I refuse to accept that as a viable solution because I would rather pass std::istream around (standards!!)
	// http://www.boost.org/doc/libs/1_52_0/doc/html/boost_asio/example/timeouts/blocking_tcp_client.cpp
	// https://svn.boost.org/trac10/ticket/13207
	// D: gdb_rsp_client_example.cpp: 58: opened socket to 127.0.0.1:1234
	// D: gdb_rsp_client_example.cpp: 115: about to create begin iterator

	D() << "about to create begin iterator" << std::endl;
	boost::spirit::istream_iterator begin( ctx.fds );

	D() << "about to create end iterator" << std::endl;
	boost::spirit::istream_iterator end;

	D() << "about to create client..." << std::endl;
	org::gdb::rsp::client client( ctx.fds, ctx.fds );

	D() << "about to create qrs.." << std::endl;

	org::gdb::rsp::QuerySupportedResponse qrs;

	D() << "about to call querySupported.." << std::endl;

	client.querySupported(
		qrs,
		org::gdb::rsp::GdbFeature({
			{ "hwbreak", "+", },
			{ "swbreak", "+", },
			{ "qRelocInsn", "+", },
		})
	);

	D() << "returned from querySupported()" << std::endl;

	for( auto & kv: qrs.feat ) {
		D() << "key: '" << kv.first << "', value: '" << kv.second << "'" << std::endl;
	}

	return 0;
}
