/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <iostream>
#include <regex>

#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>
#include <boost/spirit/include/phoenix_fusion.hpp>
#include <boost/spirit/include/phoenix_stl.hpp>
#include <boost/spirit/include/phoenix_object.hpp>
#include <boost/spirit/include/qi.hpp>
#include <client.hpp>

#include "checksum.hpp"
#include "debug.hpp"
#include "exception.hpp"

namespace org {
namespace gdb {
namespace rsp {

namespace phoenix = boost::phoenix;
namespace qi = boost::spirit::qi;
namespace ascii = boost::spirit::ascii;

typedef std::deque<std::string> token_val_t;

template <typename Iterator>
struct token_sequence
  : boost::spirit::qi::grammar<Iterator, token_val_t()>
{
	token_sequence()
	  : token_sequence::base_type( tokens )
	{
		using qi::lit;
		using qi::lexeme;
		using qi::on_error;
		using qi::fail;
		using qi::retry;
		using ascii::char_;
		using ascii::string;
		using namespace qi::labels;

		using phoenix::construct;
		using phoenix::val;

		// an overview of boost operators is here:
		// http://www.boost.org/doc/libs/1_62_0/libs/spirit/doc/html/spirit/qi/reference/operator.html
		tokens %=  tok >> *( tok );
		tok %= ( ack | nack | frame );
		ack %= char_( "+" );
		nack %= char_( "-" );
		frame %= ( sof > payload > eof );
		sof %= char_( "$%" );
		payload %= lexeme[ *( char_ - '$' - '%' - '#' ) ];
		eof %= ( char_( "#" ) > char_( "0-9a-f" ) > char_( "0-9a-f" ) );

		on_error<fail>(
			tokens,
			D()
				<< val("Error<fail>! Expecting ")
				<< _4                               // what failed?
				<< val(" here: \"")
				<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
				<< val("\"")
				<< std::endl
        );
		on_error<retry>(
			tokens,
			D()
				<< val("Error<retry>! Expecting ")
				<< _4                               // what failed?
				<< val(" here: \"")
				<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
				<< val("\"")
				<< std::endl
		);
	}

	qi::rule<Iterator, token_val_t()> tokens;
	qi::rule<Iterator, std::string()> tok, ack, nack, frame, sof, payload, eof;
};

client::client( std::istream & is, std::ostream & os )
:
	is( & is ),
	os( & os ),
	begin( is ),
	noack( false ),
	longest_payload(  0 ),
	properties( default_stub_properties )
{
	D() << "created client" << std::endl;
}

client::~client() {
	D() << "destroying client.." << std::endl;
}

std::string client::get_next_frame() {

	token_sequence<boost::spirit::istream_iterator> p;
    std::string r;

    std::deque<std::string> unchecked_received_frames;

	if ( nullptr == is ) {
		E() << "aborting response read since input stream is null" << std::endl;
		return "";
	}

	if ( received_frames.empty() ) {

		D() << "about to call qi::parse()" << std::endl;

		if ( ! boost::spirit::qi::parse( begin, end, p, unchecked_received_frames ) ) {

			D() << "Parsing failed" << std::endl;
			*os << "-" << std::flush;

		} else {

			D() << "Parsing succeeded \\o/. Received " << unchecked_received_frames.size() << " (unchecked) frames!" << std::endl;

			for( auto & f: unchecked_received_frames ) {

				D() << "(unchecked) frame is '" << f << "'" << std::endl;

				if ( false ) {
				} else if ( "+" == f ) {

					D() << "Received an ACK! \\o/" << std::endl;

					received_frames.push_back( f );

				} else if ( "-" == f ) {

					D() << "Received a NACK! :(" << std::endl;

					// in noack mode, we handle nacks on the receive side, in ack mode, we handle it on the transmit side
					if ( noack ) {
						E() << "received a NACK, resending last frame '" << last_outgoing_frame << "'" << std::endl;
						*os << last_outgoing_frame << std::flush;
					}

				} else {

					std::string cs_str = f.substr( f.length() - 2, 2 );
					f = f.substr( 1, f.length() - std::string( "$#xx" ).length() );
					D() << "frame is '" << f << "' cs is '" << cs_str << "'" << std::endl;

					uint8_t expected_cs = checksum::calculate( f );
					uint8_t actual_cs = std::stoi( cs_str, nullptr, 16 );

					if ( actual_cs != expected_cs ) {
						E()
							<< "actual_cs " << checksum::str( actual_cs )
							<< " does not match expected checksum " << checksum::str( expected_cs )
							<< std::endl;
						*os << "-" << std::flush;
						continue;
					}

					received_frames.push_back( f );
				}
			}

			D() << "Parsed " << received_frames.size() << " frames" << std::endl;
		}
    }

	if ( received_frames.empty() ) {
		r = "";
	} else {
		r = received_frames.front();
		received_frames.pop_front();
	}

	D() << "r: '" << r << "'" << std::endl;

    if ( r.length() > longest_payload ) {
    		longest_payload = r.length();
    		D() << "longest_payload: " << longest_payload << std::endl;
    }

    return r;
}

std::string client::to_packet( const std::string & payload ) {
	std::string r;
	std::stringstream ss;
	uint8_t checksum = checksum::calculate( payload );
	ss << "$" << payload << "#" << checksum::str( checksum );
	r = ss.str();
	return r;
}

void client::send_request( const std::string & request ) {

	std::string r;

	if ( nullptr == os ) {
		E() << "aborting request since output stream is null" << std::endl;
		return;
	}

	D() << "sending request '" << request << "'" << std::endl;

	r = to_packet( request );
	last_outgoing_frame = r;

	D() << "request frame is '" << last_outgoing_frame << "'" << std::endl;

	do {
		*os << last_outgoing_frame << std::flush;

		if ( ! noack ) {

			r = get_next_frame();
			if ( false ) {
			} else if ( "-" == r ) {
				// NACK -> packet was corrupted
				E() << "Received a NACK -> resending last frame" << std::endl;
				continue;
			} else if ( "" == r ) {
				throw org::gdb::rsp::exception::GdbRspUnknownPacket( last_outgoing_frame );
			} else if ( is_error_frame( r ) ) {
				throw org::gdb::rsp::exception::make( err_no( r ), last_outgoing_frame );
			} else if ( "+" == r ) {
				D() << "Received an ACK \\o/" << std::endl;
				break;
			}
		}
	} while( 0 );
}

bool client::is_error_frame( const std::string & response ) {
	return std::regex_match( response, std::regex( "E[0-9a-f][0-9a-f]" ) );
}

GdbRspError::type client::err_no( const std::string & error_response ) {
	return (GdbRspError::type) std::stoi( error_response.substr( 1 ), nullptr, 16 );
}

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */
