/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <map>

#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>
#include <boost/spirit/include/phoenix_fusion.hpp>
#include <boost/spirit/include/phoenix_stl.hpp>
#include <boost/spirit/include/phoenix_object.hpp>
#include <boost/spirit/include/qi.hpp>
#include <boost/fusion/include/std_pair.hpp>
#include <client.hpp>
#include <feature.hpp>
#include <packet.hpp>

#include "debug.hpp"

namespace org {
namespace gdb {
namespace rsp {

namespace phoenix = boost::phoenix;
namespace qi = boost::spirit::qi;
namespace ascii = boost::spirit::ascii;

typedef std::map<std::string,std::string> token_val_t;

template <typename Iterator>
struct grammar_
  : qi::grammar<Iterator, token_val_t()>
{
    grammar_()
      : grammar_::base_type( resp )
    {
		using qi::lit;
		using qi::lexeme;
		using qi::on_error;
		using qi::fail;
		using qi::retry;
		using ascii::char_;
		using ascii::string;
		using namespace qi::labels;

		using phoenix::construct;
		using phoenix::val;

        resp %= pair >> *( ';' > pair );
        pair  %= key >> ( plus | minus | ( '=' > value ) );
        key   %= lexeme[ +( char_ - '=' - '-' - '+' ) ];
        value %= lexeme[ +( char_ - ';' - '-' - '+' ) ];
        plus %= char_( "+" );
        minus %= char_( "-" );

        on_error<fail>(
        		resp,
			std::cerr
				<< val("Error<fail>! Expecting ")
				<< _4                               // what failed?
				<< val(" here: \"")
				<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
				<< val("\"")
				<< std::endl
		);
        on_error<retry>(
			resp,
			std::cerr
				<< val("Error<retry>! Expecting ")
				<< _4                               // what failed?
				<< val(" here: \"")
				<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
				<< val("\"")
				<< std::endl
		);

#ifdef DEBUG
        BOOST_SPIRIT_DEBUG_NODES(
            (resp)
        );
#endif
    }

    qi::rule<Iterator, token_val_t()> resp;
    qi::rule<Iterator, std::pair<std::string, std::string>()> pair;
    qi::rule<Iterator, std::string()> key, value, plus, minus;
};

static std::string generate( const GdbFeature & feat ) {

	std::stringstream ss;

	ss << packet_s[ packet_t::QuerySupported ];
	if ( ! feat.empty() ) {
		ss << ":";
		size_t i = 0;
		for( auto & kv: feat ) {
			ss << kv.first;
			if (
				false
				|| "+" == kv.second
				|| "-" == kv.second
				|| "?" == kv.second
			) {
				ss << kv.second;
			} else {
				// k=v
				ss << "=" << kv.second;
			}
			if ( i++ < feat.size() - 1 ) {
				ss << ";";
			}
		}
	}

	std::string r = ss.str();

	return r;
}

void client::querySupported( QuerySupportedResponse & r, const GdbFeature & arg1 ) {

	std::lock_guard<std::mutex> lock( query_mutex );

	D() << "generating request.." << std::endl;

	send_request( generate( arg1 ) );

	D() << "about to call get_next_frame()" << std::endl;

	std::string input = get_next_frame();

	D() << "input is '" << input << "'" << std::endl;

    std::string::iterator begin = input.begin();
    std::string::iterator end = input.end();

    grammar_<std::string::iterator> p;

    r.feat.clear();

    D() << "about to call parse()" << std::endl;

    if ( ! boost::spirit::qi::parse( begin, end, p, r.feat ) ) {
		E() << "failed to parse '" << input << "'" << std::endl;
		*os << to_packet( "" ) << std::flush;
    } else {
		D() << "parsed input '" << input << "'" << std::endl;

		for( auto & kv: r.feat ) {
			if ( "PacketSize" == kv.first ) {
				try {
					longest_payload = std::stoll( kv.second, nullptr, 16 ) - std::string( "$#xx" ).length();
				} catch( ... ) {
					E() << "unrecognized value '" << kv.second << "'  for " << kv.first << ". skipping" << std::endl;
					continue;
				}
			}
		}
    }
}

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */
