/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <map>
#include <tuple>
#include <vector>

#include <boost/algorithm/string.hpp>
#include <boost/fusion/adapted/struct/adapt_struct.hpp>
#include <boost/fusion/include/adapt_struct.hpp>
#include <boost/fusion/include/std_pair.hpp>
#include <boost/fusion/include/std_tuple.hpp>
#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>
#include <boost/spirit/include/phoenix_fusion.hpp>
#include <boost/spirit/include/phoenix_stl.hpp>
#include <boost/spirit/include/phoenix_object.hpp>
#include <boost/spirit/include/qi.hpp>
#include <boost/spirit/include/qi_list.hpp>

#include "client.hpp"
#include "debug.hpp"
#include "feature.hpp"
#include "packet.hpp"
#include "exception.hpp"
#include "hexstr.hpp"

namespace org {
namespace gdb {
namespace rsp {

struct triplet_s {
	triplet_s()
	:
		c( 0 )
	{
	}

	triplet_s( const std::string & a, const std::string & b, const int c )
	:
		a( a ), b( b ), c( c )
	{
	}
	std::string a;
	// in the response packet, 'text' is hex-encoded.
	// Here we give the decoded string
	std::string b;
	int c;
};

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */

BOOST_FUSION_ADAPT_STRUCT(
	org::gdb::rsp::triplet_s,
	( std::string, a )
	( std::string, b )
	( int, c )
)

namespace org {
namespace gdb {
namespace rsp {

typedef struct token_val_s token_val_t;

namespace phoenix = boost::phoenix;
namespace qi = boost::spirit::qi;
namespace ascii = boost::spirit::ascii;

template <typename Iterator>
struct doublet_grammar_
  : qi::grammar<Iterator, triplet_s()>
{
    doublet_grammar_()
      : doublet_grammar_::base_type( resp )
    {
		using qi::lit;
		using qi::lexeme;
		using qi::on_error;
		using qi::fail;
		using qi::retry;
		using ascii::char_;
		using qi::int_;
		using qi::uint_;
		using qi::eps;
		using qi::bool_;
		using ascii::string;
		using namespace qi::labels;

		using phoenix::construct;
		using phoenix::val;

        resp %=
			lexeme[ +( char_  - ';' - ':' ) ]
			> eps
			> lit( ':' )
			> int_
			;

        on_error<fail>(
        		resp,
			std::cerr
				<< val("Error<fail>! Expecting ")
				<< _4                               // what failed?
				<< val(" here: \"")
				<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
				<< val("\"")
				<< std::endl
		);
        on_error<retry>(
			resp,
			std::cerr
				<< val("Error<retry>! Expecting ")
				<< _4                               // what failed?
				<< val(" here: \"")
				<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
				<< val("\"")
				<< std::endl
		);

#ifdef DEBUG
        BOOST_SPIRIT_DEBUG_NODES(
            (resp)
        );
#endif
    }

    qi::rule<Iterator, triplet_s()> resp;
};

template <typename Iterator>
struct triplet_grammar_
  : qi::grammar<Iterator, triplet_s()>
{
    triplet_grammar_()
      : triplet_grammar_::base_type( resp )
    {
		using qi::lit;
		using qi::lexeme;
		using qi::on_error;
		using qi::fail;
		using qi::retry;
		using ascii::char_;
		using qi::int_;
		using qi::uint_;
		using qi::eps;
		using qi::bool_;
		using ascii::string;
		using namespace qi::labels;

		using phoenix::construct;
		using phoenix::val;

        resp %=
        		lexeme[ +( char_  - ';' - ':' ) ]
			> lit( ':' )
			> hexenc
			> lit( ':' )
			> int_
			;

        hexenc = +( char_( "0-9a-fA-F" ) > char_( "0-9a-fA-F" ) )
        		[
				 _val = val( boost::spirit::qi::_1 )
			]
        		;

        on_error<fail>(
        		resp,
			std::cerr
				<< val("Error<fail>! Expecting ")
				<< _4                               // what failed?
				<< val(" here: \"")
				<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
				<< val("\"")
				<< std::endl
		);
        on_error<retry>(
			resp,
			std::cerr
				<< val("Error<retry>! Expecting ")
				<< _4                               // what failed?
				<< val(" here: \"")
				<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
				<< val("\"")
				<< std::endl
		);

        BOOST_SPIRIT_DEBUG_NODES(
        		(resp)
			(hexenc)
        	);
    }

    qi::rule<Iterator, triplet_s()> resp;
    qi::rule<Iterator, std::string()> hexenc;
};

static std::string generate() {

	std::stringstream ss;

	ss << packet_s[ packet_t::QueryTraceStatus ];

	std::string r = ss.str();

	return r;
}

void client::queryTraceStatus( QueryTraceStatusResponse & r ) {

	std::lock_guard<std::mutex> lock( query_mutex );

	D()<< "generating request.." << std::endl;
	send_request( generate() );

	D()<< "about to call get_next_frame()" << std::endl;
	std::string input = get_next_frame();

	D()<< "input is '" << input << "'" << std::endl;

	const doublet_grammar_<std::string::iterator> dp;
	const triplet_grammar_<std::string::iterator> tp;

	std::vector<std::string> field;
	std::vector<std::string> subfield;

	std::string::iterator begin;
	std::string::iterator end;

	triplet_s rr;

	QueryTraceStatusResponseField qtsrf;

	bool res;

	boost::split( field, input, boost::is_any_of( ";" ) );

	for ( size_t i = 0; i < field.size(); i++ ) {
		switch (i) {
		case 0:
			if (
				!(
					true
					&& 2 == field[0].length()
					&& 'T' == field[0][0]
					&& (
						false
						|| '0' == field[0][1]
						|| '1' == field[0][1]
					)
				)
			) {
				E() << "Sequence did not begin with T<0|1>" << std::endl;
				E() << "failed to parse '" << input << "'" << std::endl;
				*os << to_packet( "" ) << std::flush;
				return;
				//throw org::gdb::rsp::exception::make( GdbRspError::GENERIC_ERROR, "Sequence did not begin with T<0|1>" );
			}

			r.running = input[ 1 ] != '0';
			r.fields.clear();

			break;

		default:

			boost::split( subfield, field[ i ], boost::is_any_of( ":" ) );

			begin = field[ i ].begin();
			end = field[ i ].end();

			res = false;

			switch( subfield.size() ) {

			case 2:

				D() << "about to call parse( \"" << field[ i ] << "\" )" << std::endl;
				res = boost::spirit::qi::parse( begin, end, dp, rr );

				break;

			case 3:

				D() << "about to call parse( \"" << field[ i ] << "\" )" << std::endl;
				res = boost::spirit::qi::parse( begin, end, tp, rr );

				break;

			default:
				E() << "Expected 2 or 3 subfields, not " << std::to_string( subfield.size() ) << std::endl;
				E() << "failed to parse '" << input << "'" << std::endl;
				*os << to_packet( "" ) << std::flush;
				return;
				//throw org::gdb::rsp::exception::make( GdbRspError::GENERIC_ERROR, "Field " + field[ i ] + " did not decompose to 2 or 3 subfields" );
				break;
			}

			if ( ! res ) {
				E() << "failed to parse field '" << field[ i ] << "'" << std::endl;
				E() << "failed to parse '" << input << "'" << std::endl;
				*os << to_packet( "" ) << std::flush;
				return;
			}

			qtsrf.field = rr.a;
			qtsrf.text = rr.b;
			qtsrf.num = rr.c;

			r.fields.push_back( qtsrf );
			D() << qtsrf << std::endl;

			break;
		}
	}

	D()<< "parsed input '" << input << "'" << std::endl;

	D()<< "running: " << r.running
	<< std::endl
	<< "fields:"
	<< std::endl;

	for (auto & f : r.fields) {
		D()<<f << std::endl;
	}
}

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */
