/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <map>

#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>
#include <boost/spirit/include/phoenix_fusion.hpp>
#include <boost/spirit/include/phoenix_stl.hpp>
#include <boost/spirit/include/phoenix_object.hpp>
#include <boost/spirit/include/qi.hpp>
#include <boost/fusion/include/std_pair.hpp>
#include <client.hpp>
#include <feature.hpp>
#include <packet.hpp>

#include "debug.hpp"
#include "exception.hpp"

namespace org {
namespace gdb {
namespace rsp {

namespace phoenix = boost::phoenix;
namespace qi = boost::spirit::qi;
namespace ascii = boost::spirit::ascii;

typedef std::pair<std::string,std::string> token_val_t;

template <typename Iterator>
struct grammar_
  : qi::grammar<Iterator, token_val_t()>
{
    grammar_()
      : grammar_::base_type( resp )
    {
		using qi::lit;
		using qi::lexeme;
		using qi::on_error;
		using qi::fail;
		using qi::retry;
		using ascii::char_;
		using ascii::string;
		using namespace qi::labels;

		using phoenix::construct;
		using phoenix::val;

        resp %= ( mdata | ldata | err );
        mdata %= char_( "m" ) >> data;
        ldata %= char_( "l" ) >> *( data );
        err %= char_( "E" ) >> ( char_( "0-9a-fA-F" ) > char_( "0-9a-fA-F" ) );
        data %= lexeme[ +( char_ - '$' - '%' - '#' ) ];

        on_error<fail>
    	(
          resp,
    	  std::cerr
    			<< val("Error<fail>! Expecting ")
    			<< _4                               // what failed?
    			<< val(" here: \"")
    			<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
    			<< val("\"")
    			<< std::endl
    	);
        on_error<retry>
    	(
          resp,
    	  std::cerr
    			<< val("Error<retry>! Expecting ")
    			<< _4                               // what failed?
    			<< val(" here: \"")
    			<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
    			<< val("\"")
    			<< std::endl
    	);
    }

    qi::rule<Iterator, token_val_t()> resp;
    qi::rule<Iterator, token_val_t()> mdata;
    qi::rule<Iterator, token_val_t()> ldata;
    qi::rule<Iterator, token_val_t()> err;
    qi::rule<Iterator, std::string()> data;
};

static std::string generate( const std::string& arg1 /* annex */ , const int32_t arg2 /* offset */, const int32_t arg3 /* length */ ) {

	std::stringstream ss;

	ss << packet_s[ packet_t::QueryXferFeaturesRead ];
	ss << ":";
	ss << std::hex << arg2;
	ss << ",";

	std::string r = ss.str();

	return r;
}

void client::queryXferFeaturesRead(QueryXferFeaturesReadResponse& _return, const std::string& arg1 /* annex */, const int32_t arg2 /* offset */, const int32_t arg3 /* length */) {

	std::lock_guard<std::mutex> lock( query_mutex );

	if ( (size_t) arg3 > longest_payload ) {
		throw exception::make( GdbRspError::GENERIC_ERROR, "length " + std::to_string( arg3 ) + " > longest allowed payload " + std::to_string( longest_payload ) );
	}

	D() << "generating request.." << std::endl;

	send_request( generate( arg1, arg2, arg3 ) );

	D() << "about to call get_next_frame()" << std::endl;

	std::string input = get_next_frame();

	D() << "input is '" << input << "'" << std::endl;

	if ( "" == input ) {
		throw exception::make( GdbRspError::NOT_SUPPORTED_ERROR, packet_s[ packet_t::QueryXferFeaturesRead ] + " not supported" );
	}

    std::string::iterator begin = input.begin();
    std::string::iterator end = input.end();

    grammar_<std::string::iterator> p;
    token_val_t rr;

    D() << "about to call parse()" << std::endl;

    if ( ! boost::spirit::qi::parse( begin, end, p, rr ) ) {
    		E() << "failed to parse '" << input << "'" << std::endl;
    		*os << "$#00" << std::flush;
    } else {
    		D() << "parsed input '" << input << "'" << std::endl;

    		switch( rr.first[ 0 ] ) {
    		case 'm':
			_return.data = rr.second;
			_return.more = true;
			break;
    		case 'l':
			_return.data = rr.second;
			_return.more = false;
			break;
    		case 'E':
			int err = std::stoi( rr.second, nullptr, 16 );
			throw exception::make( (GdbRspError::type) err, "error " + std::to_string( err ) + " encountered" );
			break;
		}

    }
}

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */
