/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <map>
#include <string>

#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>
#include <boost/spirit/include/phoenix_fusion.hpp>
#include <boost/spirit/include/phoenix_stl.hpp>
#include <boost/spirit/include/phoenix_object.hpp>
#include <boost/spirit/include/qi.hpp>
#include <boost/fusion/include/std_pair.hpp>
#include <client.hpp>
#include <feature.hpp>
#include <packet.hpp>

#include "gdbrsp_constants.h"
#include "hexstr.hpp"
#include "pairstr.hpp"
#include "vecstr.hpp"

#include "debug.hpp"

typedef std::pair<std::string,std::string> pair_ss_t;
typedef std::vector<pair_ss_t> v_pair_ss_t;

namespace org {
namespace gdb {
namespace rsp {

struct stop_reply {

	std::string type;
	std::string hexval;
	int intval;
	v_pair_ss_t data;

	stop_reply()
	:
		intval( 0 )
	{
	}

	stop_reply( const std::string & type )
	: stop_reply( type, "", -1, v_pair_ss_t() )
	{
	}

	stop_reply( const std::string & type, const std::string & hexval )
	: stop_reply( type, hexval, -1, v_pair_ss_t() )
	{
	}

	stop_reply( const std::string & type, const std::string & hexval, const int intval )
	: stop_reply( type, hexval, intval, v_pair_ss_t() )
	{
	}

	stop_reply( const std::string & type, const std::string & hexval, const int intval, const v_pair_ss_t & data )
	:
		type( type ), hexval( hexval ), intval( intval ), data( data )
	{
		if ( -1 == intval ) {
			this->intval = hexstr::get_byte( hexval );
		}
	}
};

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */

std::ostream & operator<<( std::ostream & os, const org::gdb::rsp::stop_reply & sr) {
	os << "{ "
		<< sr.type << ", "
		<< sr.hexval << ", "
		<< sr.intval << ", "
		<< vec2str( sr.data )
	<< "}";
	return os;
}

BOOST_FUSION_ADAPT_STRUCT(
	org::gdb::rsp::stop_reply,
	( std::string, type )
	( std::string, hexval )
	( int, intval )
	( v_pair_ss_t, data )
)

namespace org {
namespace gdb {
namespace rsp {

namespace phoenix = boost::phoenix;
namespace qi = boost::spirit::qi;
namespace ascii = boost::spirit::ascii;

typedef stop_reply token_val_t;

template <typename Iterator>
struct grammar_
  : qi::grammar<Iterator, token_val_t()>
{
    grammar_()
      : grammar_::base_type( resp )
    {
		using qi::lit;
		using qi::lexeme;
		using qi::on_error;
		using qi::fail;
		using qi::retry;
		using ascii::char_;
		using ascii::string;
		using namespace qi::labels;

		using phoenix::construct;
		using phoenix::val;

        resp %= (
				ok
				| sig
				| sig_info
				| proc_exit
				| proc_term
				| thread_exit
				| no_threads
				| debug_console
				| host_syscall
			)
			;

        ok = string( g_gdbrsp_constants.STOP_REPLY_OK )
        		[
				 _val = construct<stop_reply>( _1 )
			]
        		;

        sig = (
				string( g_gdbrsp_constants.STOP_REPLY_SIG )
				> hexbytestr
			)
        		[
				 _val = construct<stop_reply>( _1, _2 )
			]
        		;

        sig_info = (
				string( g_gdbrsp_constants.STOP_REPLY_SIG_INFO )
				> hexbytestr
				> mapping
			)
        		[
				 _val = construct<stop_reply>( _1, _2, -1, _3 )
			]
        		;

        hexbytestr %= ( char_( "0-9a-fA-F" ) > char_( "0-9a-fA-F" ) );
        hexstr %= +( hexbytestr );

        mapping %= stringpair % lit( ';' );

        stringpair %=
        		lexeme[ +( char_ - ':' - ';' ) ]
			> lit( ':' )
			> lexeme[ +( char_ - ':' - ';' ) ]
			;

        on_error<fail>(
        		resp,
			std::cerr
				<< val("Error<fail>! Expecting ")
				<< _4                               // what failed?
				<< val(" here: \"")
				<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
				<< val("\"")
				<< std::endl
		);
        on_error<retry>(
			resp,
			std::cerr
				<< val("Error<retry>! Expecting ")
				<< _4                               // what failed?
				<< val(" here: \"")
				<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
				<< val("\"")
				<< std::endl
		);

#ifdef DEBUG
        BOOST_SPIRIT_DEBUG_NODES(
        		(resp)
		);
#endif
    }

    qi::rule<Iterator, token_val_t()> resp;

    qi::rule<Iterator, token_val_t()> ok;
    qi::rule<Iterator, token_val_t()> sig;
    qi::rule<Iterator, token_val_t()> sig_info;
    qi::rule<Iterator, token_val_t()> proc_exit;
    qi::rule<Iterator, token_val_t()> proc_term;
    qi::rule<Iterator, token_val_t()> thread_exit;
    qi::rule<Iterator, token_val_t()> no_threads;
    qi::rule<Iterator, token_val_t()> debug_console;
    qi::rule<Iterator, token_val_t()> host_syscall;

    qi::rule<Iterator, std::string()> hexbytestr;
    qi::rule<Iterator, std::string()> hexstr;
    qi::rule<Iterator, pair_ss_t()> stringpair;
    qi::rule<Iterator, v_pair_ss_t()> mapping;
};

static std::string generate() {

	std::stringstream ss;

	ss << packet_s[ packet_t::ReportTargetStatus ];
	std::string r = ss.str();

	return r;
}

void client::reportTargetStatus( StopReplyPacket & r ) {

	std::lock_guard<std::mutex> lock( query_mutex );

	D() << "generating request.." << std::endl;

	send_request( generate() );

	D() << "about to call get_next_frame()" << std::endl;

	std::string input = get_next_frame();

	D() << "input is '" << input << "'" << std::endl;

    std::string::iterator begin = input.begin();
    std::string::iterator end = input.end();

    grammar_<std::string::iterator> p;

    //r.feat.clear();
    stop_reply rr;

    D() << "about to call parse()" << std::endl;

    if ( ! boost::spirit::qi::parse( begin, end, p, rr ) ) {
		E() << "failed to parse '" << input << "'" << std::endl;
		*os << to_packet( "" ) << std::flush;
    } else {
		D() << "parsed input '" << input << "'" << std::endl;

		r.type = rr.type;
		r.hexstr = rr.hexval;
		r.intval = rr.intval;

		for( auto & e: rr.data ) {
			r.data[ e.first ] = e.second;
		}
    }
}

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */
