/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <map>

#include <boost/spirit/include/phoenix_core.hpp>
#include <boost/spirit/include/phoenix_operator.hpp>
#include <boost/spirit/include/phoenix_fusion.hpp>
#include <boost/spirit/include/phoenix_stl.hpp>
#include <boost/spirit/include/phoenix_object.hpp>
#include <boost/spirit/include/qi.hpp>
#include <boost/fusion/include/std_pair.hpp>
#include <client.hpp>
#include <feature.hpp>
#include <packet.hpp>

#include "exception.hpp"

#include "debug.hpp"

namespace org {
namespace gdb {
namespace rsp {

namespace phoenix = boost::phoenix;
namespace qi = boost::spirit::qi;
namespace ascii = boost::spirit::ascii;

typedef std::pair<std::string,uint8_t> token_val_t;

template <typename Iterator>
struct grammar_
  : qi::grammar<Iterator, token_val_t()>
{
    grammar_()
      : grammar_::base_type( resp )
    {
		using qi::lit;
		using qi::lexeme;
		using qi::on_error;
		using qi::fail;
		using qi::retry;
		using ascii::char_;
		using ascii::string;
		using namespace qi::labels;

		using phoenix::construct;
		using phoenix::val;

        resp %= ( ok | err );
        ok = ( char_( "O" ) > char_( "K" ) )[ _val = construct<token_val_t>( val( "OK" ), val( 0 ) ) ];
        err = char_( "E" ) > (
			char_( "0-9a-fA-F" ) > char_( "0-9a-fA-F" )
			) [
			   _val = construct<token_val_t>( val( "ERR" ), val( _1 ) )
			];

        on_error<fail>
    	(
          resp,
    	  std::cerr
    			<< val("Error<fail>! Expecting ")
    			<< _4                               // what failed?
    			<< val(" here: \"")
    			<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
    			<< val("\"")
    			<< std::endl
    	);
        on_error<retry>
    	(
          resp,
    	  std::cerr
    			<< val("Error<retry>! Expecting ")
    			<< _4                               // what failed?
    			<< val(" here: \"")
    			<< construct<std::string>( _3, _2 )   // iterators to error-pos, end
    			<< val("\"")
    			<< std::endl
    	);
    }

    qi::rule<Iterator, token_val_t()> resp;
    qi::rule<Iterator, token_val_t()> ok, err;
};

static std::string generate( const SetThreadOp arg1, const ThreadId arg2 ) {

	std::stringstream ss;

	char c = (char) arg1;

	if ( ! std::isprint( c ) || '$' == c || '#' == c || '{' == c ) {
		// We intentionally only disallow absolutely unacceptable characters here in case the
		// protocol wishes to expeand. Let the GDB Stub respond with an error if something is unsupported
		throw exception::make( GdbRspError::GENERIC_ERROR, "illegal SetThreadOp character '" + std::to_string( c ) + "'" );
	}

	if ( arg2 < -1 ) {
		// technically speaking legal values are in the range [-1, 0, ... ]
		throw exception::make( GdbRspError::GENERIC_ERROR, "illegal ThreadId '" + std::to_string( arg2 ) + "'" );
	}

	ss << packet_s[ packet_t::SetThread ];
	ss << c;
	ss << arg2;

	std::string r = ss.str();

	return r;
}

void client::setThread( const SetThreadOp arg1, const ThreadId arg2 ) {

	std::lock_guard<std::mutex> lock( query_mutex );

	D() << "generating request.." << std::endl;

	send_request( generate( arg1, arg2 ) );

	D() << "about to call get_next_frame()" << std::endl;

	std::string input = get_next_frame();

	D() << "input is '" << input << "'" << std::endl;

    std::string::iterator begin = input.begin();
    std::string::iterator end = input.end();

    grammar_<std::string::iterator> p;

    token_val_t r;

    D() << "about to call parse()" << std::endl;

    if ( ! boost::spirit::qi::parse( begin, end, p, r ) ) {
    	E() << "failed to parse '" << input << "'" << std::endl;
    	*os << "$#00" << std::flush;
    } else {
    	D() << "parsed input '" << input << "'" << std::endl;

    	if ( !( "OK" == r.first && 0 == r.second ) ) {
    		throw exception::make( (GdbRspError::type) r.second, "error resulted from packet '" + this->last_outgoing_frame + "'"  );
    	}
    }
}

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */
