/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <client.hpp>
#include <iostream>
#include <sstream>
#include <string>

#include <gtest/gtest.h>

#include "client.hpp"
#include "debug.hpp"
#include "null_buffer.hpp"

namespace org {
namespace gdb {
namespace rsp {

class test_client_fixture
:
	public ::testing::Test
{

public:

	std::stringstream in;
	std::stringstream os;

	GdbFeature gdb_feature;

	test_client_fixture()
	{
		in.unsetf( std::ios_base::skipws );
	}

	virtual ~test_client_fixture()
	{
	}

protected:
	void SetUp();
	void TearDown();

	static void SetUpTestCase();
	static void TearDownTestCase();
};

void test_client_fixture::SetUpTestCase() {
}

void test_client_fixture::TearDownTestCase() {
}

void test_client_fixture::SetUp() {
}

void test_client_fixture::TearDown() {
}

TEST_F( test_client_fixture, write_to_stringstream_addr ) {
	std::string expected_string = "-";
	std::string actual_string;
	std::ostream *x = & os;

	*x << expected_string << std::flush;

	actual_string = os.str();

	EXPECT_EQ( expected_string, actual_string );
}


TEST_F( test_client_fixture, ok_frame ) {

	std::exception_ptr p = nullptr;
	std::string msg;

	std::string expected_string;
	std::string actual_string;

	size_t expected_size_t;
	size_t actual_size_t;

	in.str( "$OK#9a" );
	client client( in, os );

	expected_string = "OK";
	try {
		actual_string = client.get_next_frame();
	} catch( ... ) {
		p = std::current_exception();
	}
	EXPECT_EQ( expected_string, actual_string ) << "Did not receive correct payload";


	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown " << msg << std::endl;

	expected_size_t = expected_string.length();
	actual_size_t = client.longest_payload;
	EXPECT_EQ( expected_size_t, actual_size_t ) << "Longest payload length is incorrect";
}

TEST_F( test_client_fixture, ok_frame_async ) {

	std::exception_ptr p = nullptr;
	std::string msg;

	std::string expected_string;
	std::string actual_string;

	size_t expected_size_t;
	size_t actual_size_t;

	in.str( "%OK#9a" );
	client client( in, os );

	expected_string = "OK";
	try {
		actual_string = client.get_next_frame();
	} catch( ... ) {
		p = std::current_exception();
	}
	EXPECT_EQ( expected_string, actual_string ) << "Did not receive correct payload";


	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown " << msg << std::endl;

	expected_size_t = expected_string.length();
	actual_size_t = client.longest_payload;
	EXPECT_EQ( expected_size_t, actual_size_t ) << "Longest payload length is incorrect";
}

TEST_F( test_client_fixture, ok_frame_bad_checksum ) {

	std::string expected_string;
	std::string actual_string;

	std::string next_frame;
	std::exception_ptr p = nullptr;
	std::string msg;

	in.str( "$OK#9c" );
	client client( in, os );

	try {
		next_frame = client.get_next_frame();
	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown " << msg;

	expected_string = "-";
	actual_string = os.str();
	EXPECT_EQ( expected_string, actual_string ) << "No NACK was sent on the output stream";
}

TEST_F( test_client_fixture, bad_checksum_digits ) {

	std::string expected_string;
	std::string actual_string;

	std::string next_frame;
	std::exception_ptr p = nullptr;
	std::string msg;

	in.str( "$#sq" );
	client client( in, os );

	try {
		next_frame = client.get_next_frame();
	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown " << msg;

	expected_string = "-";
	actual_string = os.str();
	EXPECT_EQ( expected_string, actual_string ) << "No NACK was sent on the output stream";
}

TEST_F( test_client_fixture, garbage_no_reserved_tokens ) {

	std::string expected_string;
	std::string actual_string;

	std::string next_frame;
	std::exception_ptr p = nullptr;
	std::string msg;

	in.str( "asdb;lahsp13212&" );
	client client( in, os );

	try {
		next_frame = client.get_next_frame();
	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown " << msg;

	expected_string = "-";
	actual_string = os.str();
	EXPECT_EQ( expected_string, actual_string ) << "No NACK was sent on the output stream";
}

TEST_F( test_client_fixture, garbage_reserved_tokens ) {

	std::string expected_string;
	std::string actual_string;

	std::string next_frame;
	std::exception_ptr p = nullptr;
	std::string msg;

	in.str( "asdb;#asdfase$lahsp#1q#%3212&" );
	client client( in, os );

	try {
		next_frame = client.get_next_frame();
	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown " << msg;

	expected_string = "-";
	actual_string = os.str();
	EXPECT_EQ( expected_string, actual_string ) << "No NACK was sent on the output stream";
}


TEST_F( test_client_fixture, bad_checksum ) {

	std::string expected_string;
	std::string actual_string;

	std::string next_frame;
	std::exception_ptr p = nullptr;
	std::string msg;

	in.str( "$#01" );
	client client( in, os );

	try {
		next_frame = client.get_next_frame();
	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown " << msg;

	expected_string = "-";
	actual_string = os.str();
	EXPECT_EQ( expected_string, actual_string ) << "No NACK was sent on the output stream";
}


} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */
