/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <iostream>
#include <sstream>
#include <string>

#include <gtest/gtest.h>

#include "client.hpp"
#include "null_buffer.hpp"

namespace org {
namespace gdb {
namespace rsp {

class query_trace_status_fixture
:
	public ::testing::Test
{

public:

	std::ostream os;
	std::stringstream in;

	query_trace_status_fixture()
	:
		os( & null_buffer )
	{
		in.unsetf( std::ios_base::skipws );
	}

	virtual ~query_trace_status_fixture()
	{
	}

protected:
	void SetUp();
	void TearDown();

	static void SetUpTestCase();
	static void TearDownTestCase();
};

void query_trace_status_fixture::SetUpTestCase() {
}

void query_trace_status_fixture::TearDownTestCase() {
}

void query_trace_status_fixture::SetUp() {
}

void query_trace_status_fixture::TearDown() {
}

TEST_F( query_trace_status_fixture, no_fields_not_running ) {

	std::string msg;
	QueryTraceStatusResponse expected_qtsr;
	QueryTraceStatusResponse actual_qtsr;

	std::exception_ptr p = nullptr;

	std::string expected_string;
	std::string actual_string;

	bool expected_bool;
	bool actual_bool;

	size_t expected_size_t;
	size_t actual_size_t;

	in.str( "+" + client::to_packet( "T0" ) );

	client client( in, os );

	try {
		client.queryTraceStatus( actual_qtsr );

		expected_bool = false;
		actual_bool = actual_qtsr.running;
		EXPECT_EQ( expected_bool, actual_bool )
			<< "expected: " << expected_bool
			<< "actual: " << actual_bool;

		expected_size_t = 0;
		actual_size_t = actual_qtsr.fields.size();
		EXPECT_EQ( expected_size_t, actual_size_t )
			<< "expected: " << expected_size_t
			<< "actual: " << actual_size_t;

		int i = 0;
		for( auto & s: actual_qtsr.fields ) {
			std::cout << "fields[ " << i << " ]: '" << s << "'" << std::endl;
			i++;
		}

	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;
}

TEST_F( query_trace_status_fixture, no_fields_running ) {

	std::string msg;
	QueryTraceStatusResponse expected_qtsr;
	QueryTraceStatusResponse actual_qtsr;

	std::exception_ptr p = nullptr;

	std::string expected_string;
	std::string actual_string;

	bool expected_bool;
	bool actual_bool;

	size_t expected_size_t;
	size_t actual_size_t;

	in.str( "+" + client::to_packet( "T1" ) );

	client client( in, os );

	try {
		client.queryTraceStatus( actual_qtsr );

		expected_bool = true;
		actual_bool = actual_qtsr.running;
		EXPECT_EQ( expected_bool, actual_bool )
			<< "expected: " << expected_bool
			<< "actual: " << actual_bool;

		expected_size_t = 0;
		actual_size_t = actual_qtsr.fields.size();
		EXPECT_EQ( expected_size_t, actual_size_t )
			<< "expected: " << expected_size_t
			<< "actual: " << actual_size_t;

		int i = 0;
		for( auto & s: actual_qtsr.fields ) {
			std::cout << "fields[ " << i << " ]: '" << s << "'" << std::endl;
			i++;
		}

	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;
}

TEST_F( query_trace_status_fixture, doublet ) {

	std::string msg;
	QueryTraceStatusResponse expected_qtsr;
	QueryTraceStatusResponse actual_qtsr;

	std::exception_ptr p = nullptr;

	std::string expected_string;
	std::string actual_string;

	bool expected_bool;
	bool actual_bool;

	size_t expected_size_t;
	size_t actual_size_t;

	in.str( "+" + client::to_packet( "T0;tfull:0" ) );

	client client( in, os );

	try {
		client.queryTraceStatus( actual_qtsr );

		expected_bool = false;
		actual_bool = actual_qtsr.running;
		EXPECT_EQ( expected_bool, actual_bool )
			<< "expected: " << expected_bool
			<< "actual: " << actual_bool;

		expected_size_t = 1;
		actual_size_t = actual_qtsr.fields.size();
		EXPECT_EQ( expected_size_t, actual_size_t )
			<< "expected: " << expected_size_t
			<< "actual: " << actual_size_t;

		int i = 0;
		for( auto & s: actual_qtsr.fields ) {
			std::cout << "fields[ " << i << " ]: '" << s << "'" << std::endl;
			i++;
		}

	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;
}

TEST_F( query_trace_status_fixture, triplet ) {

	std::string msg;
	QueryTraceStatusResponse expected_qtsr;
	QueryTraceStatusResponse actual_qtsr;

	std::exception_ptr p = nullptr;

	std::string expected_string;
	std::string actual_string;

	bool expected_bool;
	bool actual_bool;

	size_t expected_size_t;
	size_t actual_size_t;

	in.str( "+" + client::to_packet( "T0;tstop:aabbccdd:0" ) );

	client client( in, os );

	try {
		client.queryTraceStatus( actual_qtsr );

		expected_bool = false;
		actual_bool = actual_qtsr.running;
		EXPECT_EQ( expected_bool, actual_bool )
			<< "expected: " << expected_bool
			<< "actual: " << actual_bool;

		expected_size_t = 1;
		actual_size_t = actual_qtsr.fields.size();
		EXPECT_EQ( expected_size_t, actual_size_t )
			<< "expected: " << expected_size_t
			<< "actual: " << actual_size_t;

		int i = 0;
		for( auto & s: actual_qtsr.fields ) {
			std::cout << "fields[ " << i << " ]: '" << s << "'" << std::endl;
			i++;
		}

	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;
}

TEST_F( query_trace_status_fixture, doublet_followed_by_triplet ) {

	std::string msg;
	QueryTraceStatusResponse expected_qtsr;
	QueryTraceStatusResponse actual_qtsr;

	std::exception_ptr p = nullptr;

	std::string expected_string;
	std::string actual_string;

	bool expected_bool;
	bool actual_bool;

	size_t expected_size_t;
	size_t actual_size_t;

	in.str( "+" + client::to_packet( "T0;tfull:0;tstop:aabbccdd:0" ) );

	client client( in, os );

	try {
		client.queryTraceStatus( actual_qtsr );

		expected_bool = false;
		actual_bool = actual_qtsr.running;
		EXPECT_EQ( expected_bool, actual_bool )
			<< "expected: " << expected_bool
			<< "actual: " << actual_bool;

		expected_size_t = 2;
		actual_size_t = actual_qtsr.fields.size();
		EXPECT_EQ( expected_size_t, actual_size_t )
			<< "expected: " << expected_size_t
			<< "actual: " << actual_size_t;

		int i = 0;
		for( auto & s: actual_qtsr.fields ) {
			std::cout << "fields[ " << i << " ]: '" << s << "'" << std::endl;
			i++;
		}

	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;
}

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */
