/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <iostream>
#include <sstream>
#include <string>

#include <gtest/gtest.h>

#include "client.hpp"
#include "null_buffer.hpp"

namespace org {
namespace gdb {
namespace rsp {

class query_xfer_features_read_fixture
:
	public ::testing::Test
{

public:

	std::ostream os;
	std::stringstream in;

	GdbFeature gdb_feature;

	query_xfer_features_read_fixture()
	:
		os( & null_buffer )
	{
		in.unsetf( std::ios_base::skipws );
	}

	virtual ~query_xfer_features_read_fixture()
	{
	}

protected:
	void SetUp();
	void TearDown();

	static void SetUpTestCase();
	static void TearDownTestCase();
};

void query_xfer_features_read_fixture::SetUpTestCase() {
}

void query_xfer_features_read_fixture::TearDownTestCase() {
}

void query_xfer_features_read_fixture::SetUp() {
}

void query_xfer_features_read_fixture::TearDown() {
}

TEST_F( query_xfer_features_read_fixture, target_xml ) {

	std::string msg;
	QueryXferFeaturesReadResponse resp;

	std::exception_ptr p = nullptr;

	std::string expected_string = "<?xml version=\"1.0\"?><!DOCTYPE target SYSTEM \"gdb-target.dtd\"><target><architecture>arm</architecture><xi:include href=\"arm-core.xml\"/></target>";
	std::string actual_string;

	bool expected_bool = false;
	bool actual_bool;

	in.str( "+" + client::to_packet( "l" + expected_string ) );

	client client( in, os );

	try {
		client.longest_payload = 0x1000 - 4;
		client.queryXferFeaturesRead( resp, "target.xml", 0, client.longest_payload );

		actual_string = resp.data;

		EXPECT_EQ( expected_string, actual_string );

		expected_bool = false;
		actual_bool = resp.more;

		EXPECT_EQ( expected_bool, actual_bool )
			<< "expected no more data" << std::endl;

	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;
}

TEST_F( query_xfer_features_read_fixture, no_more_data ) {

	std::string msg;
	QueryXferFeaturesReadResponse resp;

	std::exception_ptr p = nullptr;

	std::string expected_string = "";
	std::string actual_string;

	bool expected_bool;
	bool actual_bool;

	in.str( "+" + client::to_packet( "l" + expected_string ) );

	client client( in, os );

	try {
		client.longest_payload = 0x1000 - 4;
		client.queryXferFeaturesRead( resp, "target.xml", 0, client.longest_payload );

		actual_string = resp.data;

		EXPECT_EQ( expected_string, actual_string );

		expected_bool = false;
		actual_bool = resp.more;

		EXPECT_EQ( expected_bool, actual_bool )
			<< "expected no more data" << std::endl;

	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;
}

TEST_F( query_xfer_features_read_fixture, more_data ) {

	std::string msg;
	QueryXferFeaturesReadResponse resp;

	std::exception_ptr p = nullptr;

	std::string expected_string = "abcd";
	std::string actual_string;

	bool expected_bool;
	bool actual_bool;

	in.str( "+" + client::to_packet( "m" + expected_string ) );

	client client( in, os );

	try {
		client.longest_payload = 0x1000 - 4;
		client.queryXferFeaturesRead( resp, "target.xml", 0, client.longest_payload );

		actual_string = resp.data;

		EXPECT_EQ( expected_string, actual_string );

		expected_bool = true;
		actual_bool = resp.more;

		EXPECT_EQ( expected_bool, actual_bool )
			<< "expected more data" << std::endl;

	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;
}

TEST_F( query_xfer_features_read_fixture, error ) {

	std::string msg;
	QueryXferFeaturesReadResponse resp;

	std::exception_ptr p = nullptr;

	std::string expected_string = "E00";
	std::string actual_string;

	bool expected_bool;
	bool actual_bool;

	in.str( "+" + client::to_packet( expected_string ) );

	client client( in, os );

	try {
		client.longest_payload = 0x1000 - 4;
		client.queryXferFeaturesRead( resp, "target.xml", 0, client.longest_payload );

		actual_string = resp.data;

		EXPECT_EQ( expected_string, actual_string );

		expected_bool = false;
		actual_bool = resp.more;

		EXPECT_EQ( expected_bool, actual_bool )
			<< "expected no more data" << std::endl;

	} catch( ... ) {
		p = std::current_exception();
	}

	EXPECT_FALSE( nullptr == p ) << "exception was *not* thrown";
}

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */
