/*
 * The MIT License (MIT)
 *
 * Copyright (c) 2017, Christopher Friedt <chrisfriedt@gmail.com>
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in all
 * copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
 * AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <client.hpp>
#include <iostream>
#include <sstream>
#include <string>

#include <gtest/gtest.h>

#include "client.hpp"
#include "null_buffer.hpp"

#include "../src/mapstr.hpp"

namespace org {
namespace gdb {
namespace rsp {

class test_report_target_status_fixture
:
	public ::testing::Test
{

public:

	std::ostream os;
	std::stringstream in;

	GdbFeature gdb_feature;

	test_report_target_status_fixture()
	:
		os( & null_buffer )
	{
		in.unsetf( std::ios_base::skipws );
	}

	virtual ~test_report_target_status_fixture()
	{
	}

protected:
	void SetUp();
	void TearDown();

	static void SetUpTestCase();
	static void TearDownTestCase();
};

void test_report_target_status_fixture::SetUpTestCase() {
}

void test_report_target_status_fixture::TearDownTestCase() {
}

void test_report_target_status_fixture::SetUp() {
}

void test_report_target_status_fixture::TearDown() {
}

TEST_F( test_report_target_status_fixture, TestStopReplyOk ) {

	StopReplyType expected_StopReplyType;
	StopReplyType actual_StopReplyType;

	StopReplyPacket r;

	std::string msg;
	std::exception_ptr p = nullptr;

	in.str( "+$OK#9a" );

	client client( in, os );

	try {
		client.reportTargetStatus( r );
	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;

	expected_StopReplyType = g_gdbrsp_constants.STOP_REPLY_OK;
	actual_StopReplyType = r.type;

	EXPECT_EQ( expected_StopReplyType, actual_StopReplyType )
		<< "expected: " << expected_StopReplyType << " "
		<< "actual: " << actual_StopReplyType
		<< std::endl;
}

TEST_F( test_report_target_status_fixture, TestStopReplySig ) {

	StopReplyType expected_StopReplyType;
	StopReplyType actual_StopReplyType;

	std::string expected_string;
	std::string actual_string;

	int expected_int;
	int actual_int;

	StopReplyPacket r;

	std::string msg;
	std::exception_ptr p = nullptr;

	in.str( "+$S05#b8" );

	client client( in, os );

	try {
		client.reportTargetStatus( r );
	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;

	expected_StopReplyType = g_gdbrsp_constants.STOP_REPLY_SIG;
	actual_StopReplyType = r.type;

	EXPECT_EQ( expected_StopReplyType, actual_StopReplyType )
		<< "expected: " << expected_StopReplyType << " "
		<< "actual: " << actual_StopReplyType
		<< std::endl;

	expected_string = "05";
	actual_string = r.hexstr;

	EXPECT_EQ( expected_string, actual_string )
		<< "expected: " << expected_string << " "
		<< "actual: " << actual_string
		<< std::endl;

	expected_int = 5;
	actual_int = r.intval;

	EXPECT_EQ( expected_int, actual_int )
		<< "expected: " << expected_int << " "
		<< "actual: " << actual_int
		<< std::endl;
}

TEST_F( test_report_target_status_fixture, TestStopReplySigInfo ) {

	StopReplyType expected_StopReplyType;
	StopReplyType actual_StopReplyType;

	std::string expected_string;
	std::string actual_string;

	int expected_int;
	int actual_int;

	typedef std::map<std::string,std::string> mapss;
	mapss expected_mapss;
	mapss actual_mapss;

	StopReplyPacket r;

	std::string msg;
	std::exception_ptr p = nullptr;

	in.str( "+$T05thread:01;#07" );

	client client( in, os );

	try {
		client.reportTargetStatus( r );
	} catch( ... ) {
		p = std::current_exception();
	}

	if ( nullptr != p ) {
		try {
			std::rethrow_exception( p );
		} catch( std::exception & e ) {
			msg = e.what();
		}
	}

	EXPECT_TRUE( nullptr == p ) << "exception was thrown with message " << msg;

	expected_StopReplyType = g_gdbrsp_constants.STOP_REPLY_SIG_INFO;
	actual_StopReplyType = r.type;

	EXPECT_EQ( expected_StopReplyType, actual_StopReplyType )
		<< "expected: " << expected_StopReplyType << " "
		<< "actual: " << actual_StopReplyType
		<< std::endl;

	expected_string = "05";
	actual_string = r.hexstr;

	EXPECT_EQ( expected_string, actual_string )
		<< "expected: " << expected_string << " "
		<< "actual: " << actual_string
		<< std::endl;

	expected_int = 5;
	actual_int = r.intval;

	EXPECT_EQ( expected_int, actual_int )
		<< "expected: " << expected_int << " "
		<< "actual: " << actual_int
		<< std::endl;

	expected_mapss = mapss{
		{ "thread", "01" },
	};
	actual_mapss = r.data;

	EXPECT_EQ( expected_mapss, actual_mapss )
		<< "expected: " << map2str( expected_mapss ) << " "
		<< "actual: " << map2str( actual_mapss )
		<< std::endl;
}

} /* namespace rsp */
} /* namespace gdb */
} /* namespace org */
